@@ -1,6 +1,36 @@
+#! /bin/sh
+# \
+exec wish8.2 "$0" "$@"
+
+#################################################################
+# ntkinfo.tcl - a slimmed down and updated tkinfo
+# (http://math-www.uni-paderborn.de/~axel/tkinfo)
+# 
+# Copyright (C) 1999 Mark Patton
+#
+# This program is free software; you can redistribute it and/or
+# modify it under the terms of the GNU General Public License
+# as published by the Free Software Foundation; either version 2
+# of the License, or (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+###################################################################
+
+# To use as a lib, set the global var ntki_embed to a value.
+# Interface:
+# ntki::init sets up ntkinfo.
+# ntki::win ?node? pops up a new toplevel and returns it.
+
 namespace eval ntki {}
 
-proc ntki::init {args} {
+proc ntki::init {} {
     global env tcl_platform
     variable ntki
     
@@ -66,7 +96,7 @@ proc ntki::init {args} {
     bind info <Key-Down>     {%W yview scroll 1 units; break}
     bind info <Key-Up>       {%W yview scroll -1 units; break}
     
-    bind info <Key-c>        {ntki::close [winfo toplevel %W]; break}
+    bind info <Key-c>        {ntki::win_close [winfo toplevel %W]; break}
     bind info <Key-q>        {ntki::quit; break}
     
     bind info <Key-u>        {ntki::goto [winfo toplevel %W] up}
@@ -99,26 +129,20 @@ proc ntki::init {args} {
     bind info <Key-quoteright> \
 	    {ntki::prompt [winfo toplevel %W] search_backwards_regexp}
     bind info <Control-s> {ntki::search_continue [winfo toplevel %W]}
+    bind info <Control-c> {ntki::prompt_quit [winfo toplevel %W]; break}
+    bind info <Control-g> {ntki::prompt_quit [winfo toplevel %W]; break}
+    bind info <Escape>    {ntki::prompt_quit [winfo toplevel %W]}
 
     bind prompt <Control-u> {%W delete 0 end}
     bind prompt <Return>    {ntki::prompt_done [winfo toplevel %W]}
     bind prompt <Escape>    {ntki::prompt_quit [winfo toplevel %W]}
-    bind prompt <Control-g> {
-	ntki::interrupt [winfo toplevel %W]
-	ntki::prompt_quit [winfo toplevel %W]
-    }
+    bind prompt <Control-g> {ntki::prompt_quit [winfo toplevel %W]}
     bind prompt <Key-Up>    {ntki::prompt_hist_up [winfo toplevel %W]}
     bind prompt <Key-Down>  {ntki::prompt_hist_down [winfo toplevel %W]}
     bind prompt <Tab>       {ntki::prompt_complete [winfo toplevel %W]; break}
-    bind prompt <Control-c> {ntki::interrupt [winfo toplevel %W]}
+    bind prompt <Control-c> {ntki::prompt_quit [winfo toplevel %W]}
 
-    return
-}
-
-proc ntki::interrupt {w} {
-    variable ntki
-
-    set ntki(win,$w,interrupt) 1
+    rename init ""
     return
 }
 
@@ -175,7 +199,7 @@ proc ntki::win {{node ""}} {
     toplevel $w -class ntkinfo
     wm title $w Ntkinfo
     
-    wm protocol $w WM_DELETE_WINDOW [list ntki::close $w]
+    wm protocol $w WM_DELETE_WINDOW [list ntki::win_close $w]
     
     menu $w.menubar
     $w configure -menu $w.menubar
@@ -193,7 +217,7 @@ proc ntki::win {{node ""}} {
     $w.menubar.file add command -label "New Window " -accelerator w\
 	    -command "ntki::win (\$ntki::ntki(win,$w,file))\$ntki::ntki(win,$w,loc)"
     $w.menubar.file add command -label "Close Window" \
-	    -command [list ntki::close $w] -accelerator c
+	    -command [list ntki::win_close $w] -accelerator c
     $w.menubar.file add command -label Quit -command {ntki::quit} \
 	    -accelerator q   
 
@@ -209,35 +233,35 @@ proc ntki::win {{node ""}} {
 	    -command [list ntki::goto $w top]
 
     menu $w.menubar.dir
-    
+
+    set ntki(win,$w,dir) [lindex $ntki(dirs) 0]    
     foreach dir $ntki(dirs) {
 	    $w.menubar.dir add radiobutton -label $dir -value $dir\
-		    -variable ntki::tki(win,$w,dir) \
+		    -variable ntki::ntki(win,$w,dir) \
 		    -command [list ntki::goto $w dir]
     }
-    set ntki(win,$w,dir) [lindex $ntki(dirs) 0]
 
     menu $w.menubar.search
 
     $w.menubar.search add command -label "Exact forward..." -accelerator /\
-	    -command {}
+	    -command [list ntki::prompt $w search_forwards_exact]
     $w.menubar.search add command -label "Exact backward..." -accelerator \\ \
-	    -command {}
-    $w.menubar.search add command -label "Regexp forward..." \
-	    -command {}
-    $w.menubar.search add command -label "Regexp backward..." \
-	    -command {}
+	    -command [list ntki::prompt $w search_backwards_exact]
+    $w.menubar.search add command -label "Regexp forward..." -accelerator \; \
+	    -command [list ntki::prompt $w search_forwards_regexp]
+    $w.menubar.search add command -label "Regexp backward..."  -accelerator ' \
+	    -command [list ntki::prompt $w search_backwards_regexp]
     $w.menubar.search add command -label "Continue" -accelerator C-s \
-	    -command {}
+	    -command [list ntki::search_continue $w]
 
     menu $w.menubar.history
 
     menu $w.menubar.help
 
     $w.menubar.help add command -label "Contents" \
-	    -command {win_display %W (ntkinfo)Top}
+	    -command [list ntki::win_display $w (ntkinfo)Top]
     $w.menubar.help add command -label "About ntkinfo" \
-	    -command {win_display %W (ntkinfo)about}
+	    -command [list ntki::win_display $w (ntkinfo)About]
     
     frame $w.toolbar
 
@@ -326,10 +350,11 @@ proc ntki::get_cur_node {wt tag} {
     return $node
 }
 
-proc ntki::close {w} {
+proc ntki::win_close {w} {
     variable ntki
 
-    interrupt $w
+    set ntki(win,$w,interrupt) 1
+    set file $ntki(win,$w,file)
 
     foreach el [array names ntki win,$w,*] {
 	unset ntki($el)
@@ -339,7 +364,8 @@ proc ntki::close {w} {
     if {[llength [winfo children .]] == 0} {
 	exit
     }
-    
+
+    check_flush $file
     return
 }
 
@@ -350,24 +376,14 @@ proc ntki::quit {} {
     if {![info exist ntki_embed]} {
 	exit
     }
-
-    foreach win [lmatch_prefix [winfo children .] .ntki] {
-	close $win
+    
+    foreach w [lmatch_prefix [winfo children .] .ntki] {
+	win_close $w
     }
     
     return
 }
 
-proc ntki::flush {file} {
-    variable ntki
-    
-    foreach el [array names ntki info,$file,*] {
-	unset ntki($file)
-    }
-
-    return
-}
-
 # For xrefs, there are two forms:
 #   *note nodeSpec::terminator
 #   *note label: nodeSpec terminator
@@ -377,6 +393,12 @@ proc ntki::flush {file} {
 #   * nodeSpec::	comments...
 #   * label: nodeSpec[\t.,] comments...	
 
+# An info file is stored as ntki(info,$file,$node)
+# A node is a list of {file up prev next} followed
+# by {text tag} pairs.
+
+# If only on element is present then it is the file
+# we have to load for the indirect node.
 
 proc ntki::info_load {w file {to_open ""}} {
     variable ntki
@@ -523,7 +545,7 @@ proc ntki::complete {prefix {compress_only 0}} {
     variable ntki
 
     if {!$compress_only} {
-	set suf_list $ntki(info_suffix)
+	set suf_list $ntki(info_suffix) 
     }
     lappend suf_list ""
 
@@ -540,12 +562,37 @@ proc ntki::complete {prefix {compress_only 0}} {
     return
 }
 
+# flush a file if its not loaded into a window
+
+proc ntki::check_flush {file} {
+    variable ntki
+
+    set used 0
+    foreach id [array names ntki win,*,file] {
+	if {[string equal $ntki($id) $file]} {
+	    set used 1
+	    break
+	}
+    }
+    
+    if {!$used} {
+	foreach el [array names ntki info,$file,*] {
+	    unset ntki($el)
+	}
+    }
+    
+    return
+}
+
 proc ntki::win_display {w node} {
     variable ntki
 
     parse_node $w $node file loc
 
+    $w.status.msg configure -text ""
+
     if {![info exists ntki(info,$file,Top)]} {
+	after idle [list ntki::check_flush $ntki(win,$w,file)]
     	info_load $w $file
     }
 
@@ -563,8 +610,8 @@ proc ntki::win_display {w node} {
     $w.main.text configure -state normal
     $w.main.text delete 1.0 end
 
-    foreach {text tags} [lrange $ntki(info,$file,$loc) 4 end] {
-	$w.main.text insert end $text $tags
+    foreach {text tags text2 tags2} [lrange $ntki(info,$file,$loc) 4 end] {
+	$w.main.text insert end $text $tags $text2 $tags2
     }
 
     $w.main.text configure -state disabled
@@ -649,35 +696,36 @@ proc ntki::prompt {w type} {
 proc ntki::hist_add {w line} {
     variable ntki
     
-    set ntki(win,$w,history) [lreplace $ntki(win,$w,history) 0 \
-	    [expr {[llength $ntki(win,$w,history)] - $ntki(phist_len)}]]
-    lappend ntki(win,$w,history) $line
-    set ntki(win,$w,hist_loc) [llength $ntki(win,$w,history)]    
 
     return
 }
 
+# Somtimes a node's next field points to the first node in
+# a menu which can cause cycles if you don't catch the case
+
 proc ntki::lnext {w loc} {
     variable ntki
 
     set file $ntki(win,$w,file)
-    set node [lindex [nodes $w $loc menu] 0]
+    set lnext [lindex [nodes $w $loc menu] 0]
     
-    if {[string equal $node ""]} {
-        set node [lindex $ntki(info,$file,$loc) 3]
-        
-        while {[string equal $node ""]} {
-            set loc [lindex $ntki(info,$file,$loc) 1]
+    if {[string equal $lnext ""]} {
+	set lnext $loc
+
+        while {[string equal [lindex $ntki(info,$file,$lnext) 3] ""]} {
+            set lnext [lindex $ntki(info,$file,$lnext) 1]
             
-            if {[string equal $loc ""] || [string equal $loc Top]} {
-                break
-            }
-            
-            set node [lindex $ntki(info,$file,$loc) 3]
+            if {[string equal $lnext ""] || [string equal $lnext Top]} {
+                return
+            } elseif {[string equal [lindex $ntki(info,$file,$lnext) 3] \
+		    [lindex [nodes $w $lnext menu] 0]]} {
+		set lnext [lindex $ntki(info,$file,$lnext) 1]
+	    }
         }
+	set lnext [lindex $ntki(info,$file,$lnext) 3]
     }
 
-    return $node
+    return $lnext
 }
 
 proc ntki::lprev {w loc} {
@@ -699,6 +747,10 @@ proc ntki::lprev {w loc} {
     return $node
 }
 
+# do an initial search at the current location for
+# the pattern. If not found loop through the logical
+# structure of the document.
+
 proc ntki::win_search {w pattern dir type {start @0,0}} {
     variable ntki
     
@@ -709,6 +761,11 @@ proc ntki::win_search {w pattern dir type {start @0,0}} {
 	set stop 1.0
     }
 
+    if {[string equal $type regexp] && [catch {regexp $pattern ""} error]} {
+	$w.status.msg configure -text $error
+	return
+    }
+
     set beg [$wt search -$dir -$type -nocase -count c -- $pattern $start $stop]
     set ranges [$wt tag ranges sel]
 
@@ -725,6 +782,7 @@ proc ntki::win_search {w pattern dir type {start @0,0}} {
 	set ntki(win,$w,search_type) $type
 	set ntki(win,$w,search_dir) $dir
 	
+	$w.status.msg configure -text "Searching... found"
 	return
     }
 
@@ -733,19 +791,19 @@ proc ntki::win_search {w pattern dir type {start @0,0}} {
     catch {unset ntki(win,$w,interrupt)}
 
     while {1} {
-	if {[info exists ntki(win,$w,interrupt)]} {
-	    return
-	}
-
 	$w.status.msg configure -text "Searching..."
 
+	if {[info exists ntki(win,$w,interrupt)]} {
+	    $w.status.msg configure -text "Search interrupted"
+	}
+
 	if {[string equal $dir forwards]} {
 	    set loc [lnext $w $loc]
 	} else {
 	    set loc [lprev $w $loc]
 	}
-	
-	if {[string equal $loc ""]} {
+
+	if {[string equal $loc ""] || [string match (*) $loc]} {
 	    $w.status.msg configure -text "Not found"
 	    return
 	}
@@ -781,10 +839,17 @@ proc ntki::win_search {w pattern dir type {start @0,0}} {
     return
 }
 
+# cleanup after a prompt is used
+
 proc ntki::prompt_quit {w} {
+    variable ntki
+
     pack forget $w.status.input
     pack configure $w.status.msg -expand 1
     focus $w.main.text 
+    $w.status.msg configure -text ""
+
+    set ntki(win,$w,interrupt) 1
 
     return
 }
@@ -808,16 +873,18 @@ proc ntki::prompt_complete {w} {
 
     set nodes [lmatch_prefix $nodes $prefix]
 
-    if {[llength $nodes] != 1} {
+    if {[llength $nodes] == 0} {
 	bell
-    } else {
+    } elseif {[llength $nodes] == 1} {
 	$w.status.input delete 0 end
 	$w.status.input insert 0 [lindex $nodes 0]
     }
-
+    
     return
 }
 
+# called when user done entering input into prompt
+
 proc ntki::prompt_done {w} {
     variable ntki
 
@@ -826,11 +893,16 @@ proc ntki::prompt_done {w} {
     focus $w.main.text
 
     set input [$w.status.input get]
+    prompt_quit $w
+
     if {[string equal $input ""]} {
 	return
     }
 
-    hist_add $w $input
+    set ntki(win,$w,history) [lreplace $ntki(win,$w,history) 0 \
+	    [expr {[llength $ntki(win,$w,history)] - $ntki(phist_len)}]]
+    lappend ntki(win,$w,history) $input
+    set ntki(win,$w,hist_loc) [llength $ntki(win,$w,history)]    
 
     switch -glob -- $ntki(win,$w,prompt) {
 	search* {
@@ -857,9 +929,12 @@ proc ntki::prompt_done {w} {
 
     $w.status.input delete 0 end
 
+
     return
 }
 
+# return a list of all menu or xref nodes currently displayed
+
 proc ntki::nodes {w loc type} {
     variable ntki
 
@@ -875,11 +950,13 @@ proc ntki::nodes {w loc type} {
     return $nodes
 }
 
+# return a list of all nodes present in a loaded file
+
 proc ntki::all_nodes {file} {
     variable ntki
 
     set nodes [list]
-    foreach node [lmatch_prefix [array names ntki] info,$file] {
+    foreach node [array names ntki info,$file,*] {
 	lappend nodes [lindex [split $node ,] 2]
     }
     
@@ -900,20 +977,8 @@ proc ntki::search_continue {w} {
     return
 }
 
-
-proc ntki::lmatch_prefix {list str} {
-    set matches [list]
-
-    foreach el $list {
-	if {[string first $str $el] == 0} {
-	    lappend matches $el
-	}
-    }
-
-    return $matches
-}
-
 if {![info exists ntki_embed]} {
+    source ./misc.tcl
     wm withdraw .
     option add *linkColor blue
     option add *titleFont {Courier 14 bold}
